from __future__ import annotations

from collections import ChainMap
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Mapping, MutableMapping

from .typing import TEntity

if TYPE_CHECKING:
    from .fn.record import FnImplement, FnRecord


class CollectContext:
    fn_implements: dict[FnImplement, FnRecord]

    def __init__(self):
        self.fn_implements = {}

    def collect(self, entity: TEntity) -> TEntity:
        return entity.collect(self)

    @contextmanager
    def collect_scope(self):
        from .globals import COLLECTING_CONTEXT_VAR

        token = COLLECTING_CONTEXT_VAR.set(self)
        try:
            yield self
        finally:
            COLLECTING_CONTEXT_VAR.reset(token)

    @contextmanager
    def lookup_scope(self):
        from .globals import LOOKUP_LAYOUT_VAR

        token = LOOKUP_LAYOUT_VAR.set((self, *LOOKUP_LAYOUT_VAR.get()))
        try:
            yield self
        finally:
            LOOKUP_LAYOUT_VAR.reset(token)


class InstanceContext:
    instances: MutableMapping[type, Any]

    def __init__(self):
        self.instances = {}

    def store(self, *collection_or_target: Mapping[type, Any] | Any):
        for item in collection_or_target:
            if isinstance(item, Mapping):
                self.instances.update(item)
            else:
                self.instances[item.__class__] = item

    @contextmanager
    def scope(self, *, inherit: bool = True):
        from .globals import INSTANCE_CONTEXT_VAR

        if inherit:
            res = InstanceContext()
            res.instances = ChainMap({}, self.instances, INSTANCE_CONTEXT_VAR.get().instances)

            with res.scope(inherit=False):
                yield self
        else:
            token = INSTANCE_CONTEXT_VAR.set(self)
            try:
                yield self
            finally:
                INSTANCE_CONTEXT_VAR.reset(token)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Type

from .fn.overload import FnOverload


@dataclass(eq=True, frozen=True)
class SimpleOverloadSignature:
    value: Any


class SimpleOverload(FnOverload[SimpleOverloadSignature, Any, Any]):
    def digest(self, collect_value: Any) -> SimpleOverloadSignature:
        return SimpleOverloadSignature(collect_value)

    def collect(self, scope: dict, signature: SimpleOverloadSignature) -> dict[Callable, None]:
        if signature.value not in scope:
            target = scope[signature.value] = {}
        else:
            target = scope[signature.value]

        return target

    def harvest(self, scope: dict, call_value: Any) -> dict[Callable, None]:
        if call_value in scope:
            return scope[call_value]

        return {}

    def access(self, scope: dict, signature: SimpleOverloadSignature) -> dict[Callable, None] | None:
        if signature.value in scope:
            return scope[signature.value]


@dataclass(eq=True, frozen=True)
class TypeOverloadSignature:
    type: type[Any]


class TypeOverload(FnOverload[TypeOverloadSignature, Type[Any], Any]):
    def digest(self, collect_value: type) -> TypeOverloadSignature:
        return TypeOverloadSignature(collect_value)

    def collect(self, scope: dict, signature: TypeOverloadSignature) -> dict[Callable, None]:
        if signature.type not in scope:
            target = scope[signature.type] = {}
        else:
            target = scope[signature.type]

        return target

    def harvest(self, scope: dict, call_value: Any) -> dict[Callable, None]:
        t = type(call_value)
        if t in scope:
            return scope[t]

        return {}

    def access(self, scope: dict, signature: TypeOverloadSignature) -> dict[Callable, None] | None:
        if signature.type in scope:
            return scope[signature.type]


class _SingletonOverloadSignature: ...


SINGLETON_SIGN = _SingletonOverloadSignature()


class SingletonOverload(FnOverload[_SingletonOverloadSignature, None, None]):
    SIGNATURE = SINGLETON_SIGN

    def digest(self, collect_value) -> _SingletonOverloadSignature:
        return SINGLETON_SIGN

    def collect(self, scope: dict, signature) -> dict[Callable, None]:
        s = scope[None] = {}
        return s

    def harvest(self, scope: dict, call_value) -> dict[Callable, None]:
        return scope[None]

    def access(self, scope: dict, signature) -> dict[Callable, None] | None:
        if None in scope:
            return scope[None]


SINGLETON_OVERLOAD = SingletonOverload("singleton")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Generator, Generic, Protocol, TypeVar, overload

from typing_extensions import Concatenate, Self

from flywheel.globals import iter_layout

from ..typing import CQ, K1, P1, P2, C, CnQ, CnR, P, R, T
from .harvest import FnHarvestControl
from .implement import FnImplementEntity
from .record import FnImplement, FnOverloadSignal

CollectEndpointTarget = Generator[FnOverloadSignal, None, T]

A = TypeVar("A")
B = TypeVar("B", contravariant=True)


class EndpointCollectReceiver(Protocol[CnR]):
    @overload
    def __call__(self: EndpointCollectReceiver[None], entity: C) -> FnImplementEntity[C]: ...
    @overload
    def __call__(self: EndpointCollectReceiver[C], entity: FnImplementEntity[C]) -> FnImplementEntity[C]: ...
    @overload
    def __call__(self: EndpointCollectReceiver[C], entity: C) -> FnImplementEntity[C]: ...


class FnCollectEndpointAgent(Generic[P, CnQ, A, B]):
    endpoint: FnCollectEndpoint[P, CnQ]
    referrer_instance: A
    referrer_owner: type[B]

    def __init__(self, endpoint: FnCollectEndpoint[P, CnQ], referrer_instance: A, referrer_owner: type[B]):
        self.endpoint = endpoint
        self.referrer_instance = referrer_instance
        self.referrer_owner = referrer_owner

    @overload
    def __call__(
        self: FnCollectEndpointAgent[Concatenate[type[K1], P1], CQ, None, type[K1]], *args: P1.args, **kwargs: P1.kwargs
    ) -> EndpointCollectReceiver[CQ]: ...

    @overload
    def __call__(
        self: FnCollectEndpointAgent[Concatenate[K1, P1], CQ, K1, type], *args: P1.args, **kwargs: P1.kwargs
    ) -> EndpointCollectReceiver[CQ]: ...

    def __call__(self, *args, **kwargs) -> EndpointCollectReceiver:
        def receiver(entity: C | FnImplementEntity[C]) -> FnImplementEntity[C]:
            if not isinstance(entity, FnImplementEntity):
                entity = FnImplementEntity(entity)

            entity.add_target(
                self.endpoint, self.endpoint.target.__get__(self.referrer_instance, self.referrer_owner)(*args, **kwargs)
            )
            return entity

        return receiver

    def get_control(self: FnCollectEndpointAgent[..., C, Any, Any]) -> FnHarvestControl[C]:
        sig = self.endpoint.signature

        for i in iter_layout(self.endpoint):
            if sig in i.fn_implements:
                record = i.fn_implements[sig]
                break
        else:
            raise NotImplementedError(f"Cannot find record for {sig!r} in {self.endpoint!r}")

        return FnHarvestControl(self.endpoint, record)


@dataclass(init=False, eq=True, unsafe_hash=True)
class FnCollectEndpoint(Generic[P, CnQ]):
    target: Callable[P, CollectEndpointTarget]

    @overload
    def __init__(self: FnCollectEndpoint[P1, Callable[P2, R]], target: Callable[P1, CollectEndpointTarget[Callable[P2, R]]]): ...

    @overload
    def __init__(self: FnCollectEndpoint[P1, Callable], target: Callable[P1, CollectEndpointTarget[Any]]): ...

    def __init__(self, target):
        self.target = target

    @property
    def signature(self):
        return FnImplement(self)

    @overload
    def __get__(self, instance: FnCollectEndpointAgent, owner: Any) -> Self: ...

    @overload
    def __get__(self, instance: A, owner: type[B]) -> FnCollectEndpointAgent[P, CnQ, A, type[B]]: ...

    def __get__(self, instance, owner):
        if isinstance(instance, FnCollectEndpointAgent):
            return self

        return FnCollectEndpointAgent(self, instance, owner)

    def __call__(self: FnCollectEndpoint[P1, CQ], *args: P1.args, **kwargs: P1.kwargs) -> EndpointCollectReceiver[CQ]:
        def receiver(entity: C | FnImplementEntity[C]) -> FnImplementEntity[C]:
            if not isinstance(entity, FnImplementEntity):
                entity = FnImplementEntity(entity)

            entity.add_target(self, self.target(*args, **kwargs))
            return entity

        return receiver

    def get_control(self: FnCollectEndpoint[..., C]) -> FnHarvestControl[C]:
        sig = self.signature

        for i in iter_layout(self):
            if sig in i.fn_implements:
                record = i.fn_implements[sig]
                break
        else:
            raise NotImplementedError  # TODO

        return FnHarvestControl(self, record)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from ..context import CollectContext
from ..entity import BaseEntity
from ..typing import CR, P
from .record import FnRecord

if TYPE_CHECKING:
    from .endpoint import CollectEndpointTarget, FnCollectEndpoint


class FnImplementEntity(Generic[CR], BaseEntity):
    targets: list[tuple[FnCollectEndpoint, CollectEndpointTarget]]
    impl: CR

    def __init__(self, impl: CR):
        self.targets = []
        self.impl = impl

    def add_target(self, endpoint: FnCollectEndpoint[P, Any], generator: CollectEndpointTarget):
        self.targets.append((endpoint, generator))

    def collect(self, collector: CollectContext):
        super().collect(collector)

        for endpoint, generator in self.targets:
            record_signature = endpoint.signature

            if record_signature in collector.fn_implements:
                record = collector.fn_implements[record_signature]
            else:
                record = collector.fn_implements[record_signature] = FnRecord()

            for signal in generator:
                signal.overload.lay(record, signal.value, self.impl)

        return self

    def _call(self, *args, **kwargs):
        return self.impl(*args, **kwargs)

    @property
    def __call__(self) -> CR:
        return self._call  # type: ignore

    @property
    def super(self) -> CR:
        return self.targets[0][0].fn  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, Union

from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from .entity import BaseEntity


T = TypeVar("T")
R = TypeVar("R", covariant=True)
Q = TypeVar("Q", contravariant=True)
K = TypeVar("K")
K1 = TypeVar("K1")

C = TypeVar("C", bound=Callable)
CQ = TypeVar("CQ", contravariant=True, bound=Callable)
CR = TypeVar("CR", covariant=True, bound=Callable)
CnQ = TypeVar("CnQ", contravariant=True, bound=Union[Callable, None])
CnR = TypeVar("CnR", covariant=True, bound=Union[Callable, None])

P = ParamSpec("P")
P1 = ParamSpec("P1")
P2 = ParamSpec("P2")

TEntity = TypeVar("TEntity", bound="BaseEntity")


class Collectable(Protocol[P]):
    def collect(self, *args: P.args, **kwargs: P.kwargs) -> Any: ...



from .endpoint import FnCollectEndpoint as FnCollectEndpoint
from .harvest import FnHarvest as FnHarvest
from .harvest import FnHarvestControl as FnHarvestControl
from .implement import FnImplementEntity as FnImplementEntity
from .overload import FnOverload as FnOverload
from .record import FnRecord as FnRecord
